@@ -72,10 +72,15 @@ configuration file, for example
 The configuration must be in JSON format. I.e., something like
 
   {
-    "CORS": "*",
-    "Content-Type": "text/xml; charset=utf-8",
-    "version": "1.1.0",
-    "TARGET_NAMESPACE": "http://ogr.maptools.org/"
+    Common: {
+        "CORS": "*",
+        "Content-Type": "text/xml; charset=utf-8",
+        "version": "1.1.0",
+        "TARGET_NAMESPACE": "http://ogr.maptools.org/"
+    },
+    WFS: {
+        ...WFS configuration...
+    }
   }
 
 The keys and structure of this file depend on the type of the
@@ -109,7 +114,7 @@ use parent qw/Plack::Component/;
 
 binmode STDERR, ":utf8"; 
 
-our $VERSION = '0.08';
+our $VERSION = '0.09';
 
 =pod
 
@@ -138,13 +143,68 @@ sub new {
         my @json = <$fh>;
         close $fh;
         $self->{config} = decode_json "@json";
-        $self->{config}{debug} = 0 unless defined $self->{config}{debug};
+        expand_config($self->{config});
+        $self->{config}{debug} //= 0;
     }
     croak "A configuration file is needed." unless $self->{config};
     croak "No services are defined." unless $self->{services};
     return bless $self, $class;
 }
 
+sub expand_config {
+    my $config = shift;
+    my $had_ref;
+    do {
+        $had_ref = 0;
+        for my $j (keys %$config) {
+            $had_ref += config_ref($config->{$j}, $config);
+        }
+    } while ($had_ref);
+    for my $j (keys %$config) {
+        next if $j eq 'Common';
+        next unless ref $config->{$j} eq 'HASH';
+        for my $c (keys %{$config->{Common}}) {
+            $config->{$j}{$c} //= clone($config->{Common}{$c});
+        }
+    }
+}
+
+sub config_ref {
+    my ($config, $refs) = @_;
+    my $had_ref = 0;
+    if (ref $config eq 'ARRAY') {
+        for my $j (@$config) {
+            $had_ref += config_ref($j, $refs);
+        }
+    }
+    elsif (ref $config eq 'HASH') {
+        for my $j (keys %$config) {
+            my $r = $config->{$j};
+            if (ref $r) {
+                $had_ref += config_ref($r, $refs);
+            } else {
+                if ($r =~ /^ref:/) {
+                    my $target = config_target($refs, $r);
+                    croak "config reference not found: '$r'." unless $target;
+                    $config->{$j} = clone($target);
+                    $had_ref = 1;
+                }
+            }
+        }
+    }
+    return $had_ref;
+}
+
+sub config_target {
+    my ($config, $ref) = @_;
+    my @path = split /\//, $ref;
+    shift @path;
+    if (ref $config eq 'HASH') {
+        return $config->{$path[0]};
+    }
+    return undef;
+}
+
 =pod
 
 =head3 call
@@ -172,17 +232,7 @@ sub call {
 =head3 respond
 
 This method is called for each request from the Internet. The call is
-responded during the execution of the subroutine. If the request is
-"OPTIONS" the call is responded with the following headers
-
-  Content-Type = text/plain
-  Access-Control-Allow-Origin = ""
-  Access-Control-Allow-Methods = "GET,POST"
-  Access-Control-Allow-Headers = "origin,x-requested-with,content-type"
-  Access-Control-Max-Age = 60*60*24
-
-The values of Access-Control-* keys can be set in the
-configuration file. Above are the default ones.
+responded during the execution of the subroutine.
 
 In the default case this method constructs a new service object using
 the method 'service' and calls its process_request method with PSGI
@@ -196,25 +246,8 @@ processing the request.
 sub respond {
     my ($self, $responder, $env) = @_;
     if ($env->{REQUEST_METHOD} eq 'OPTIONS') {
-        my %cors = ( 
-            'Content-Type' => 'text/plain',
-            'Allow-Origin' => "",
-            'Allow-Methods' => "GET,POST",
-            'Allow-Headers' => "origin,x-requested-with,content-type",
-            'Max-Age' => 60*60*24 );
-        if (ref $self->{config}{CORS} eq 'HASH') {
-            for my $key (keys %cors) {
-                $cors{$key} = $self->{config}{CORS}{$key} // $cors{$key};
-            }
-        } else {
-            $cors{Origin} = $self->{config}{CORS};
-        }
-        for my $key (keys %cors) {
-            next if $key =~ /^Content/;
-            $cors{'Access-Control-'.$key} = $cors{$key};
-            delete $cors{$key};
-        }
-        $responder->([200, [%cors]]);
+        $responder->([200, ['Content-Type' => 'text/plain', 
+                            Geo::OGC::Service::Common::CORS($self)]]);
     } else {
         my $service;
         eval {
@@ -413,6 +446,51 @@ use Modern::Perl;
 
 =pod
 
+=head3 CORS
+
+Return the CORS headers as a list according to the configuration. CORS
+may be in the configuration as a scalar or as a hash. A scalar value
+is taken as a value for Access-Control-Allow-Origin. A hash may have
+the following keys. (Note the missing prefix Access-Control-.)
+
+      key                      default value
+  -----------------  ----------------------------------
+  Allow-Origin              
+  Allow-Credentials
+  Expose-Headers
+  Max-Age                        60*60*24
+  Allow-Methods                  GET,POST
+  Allow-Headers    origin,x-requested-with,content-type
+
+=cut
+
+sub CORS {
+    my $self = shift;
+    # default CORS response headers:
+    my %default = ( 
+        'Allow-Origin' => '',
+        'Allow-Credentials' => '',
+        'Expose-Headers' => '',
+        'Max-Age' => 60*60*24,
+        'Allow-Methods' => 'GET,POST',
+        'Allow-Headers' => 'origin,x-requested-with,content-type'
+        );
+    # where CORS is in the configuration
+    my $config = $self->{config}{Common}{CORS} // $self->{config}{CORS};
+    my @cors;
+    if (ref $config eq 'HASH') {
+        for my $key (keys %default) {
+            my $val = $config->{$key} // $default{$key};
+            push @cors, ('Access-Control-'.$key => $val);
+        }
+    } else {
+        @cors = ('Access-Control-Allow-Origin' => $config);
+    }
+    return @cors;
+}
+
+=pod
+
 =head3 DescribeService($writer)
 
 Create ows:ServiceIdentification and ows:ServiceProvider elements.
@@ -519,6 +597,7 @@ If $attribute{$key} is undefined the attribute is not written at all.
 
 package Geo::OGC::Service::XMLWriter;
 use Modern::Perl;
+use Encode qw(decode encode is_utf8);
 
 sub element {
     my $self = shift;
@@ -540,7 +619,11 @@ sub element {
     $self->write("<$element");
     if ($attributes) {
         for my $a (keys %$attributes) {
-            $self->write(" $a=\"$attributes->{$a}\"") if defined $attributes->{$a};
+            my $attr = $attributes->{$a};
+            if (defined $attr) {
+                $attr = decode utf8 => $attr unless is_utf8($attr);
+                $self->write(" $a=\"$attr\"");
+            }
         }
     }
     unless (defined $content) {
@@ -558,6 +641,7 @@ sub element {
             $self->write("</$element>");
         } elsif ($content eq '>') {
         } else {
+            $content = decode utf8 => $content unless is_utf8($content);
             $self->write("$content</$element>");
         }
     }
@@ -573,7 +657,11 @@ sub open_element {
     $self->write("<$element");
     if ($attributes) {
         for my $a (keys %$attributes) {
-            $self->write(" $a=\"$attributes->{$a}\"");
+            my $attr = $attributes->{$a};
+            if (defined $attr) {
+                $attr = decode utf8 => $attr unless is_utf8($attr);
+                $self->write(" $a=\"$attr\"");
+            }
         }
     }
     $self->write(">");
@@ -595,13 +683,13 @@ A helper class for writing XML into a stream.
 
 =head2 SYNOPSIS
 
-  my $w = Geo::OGC::Service::XMLWriter::Streaming($responder, $content_type, $declaration);
+  my $w = Geo::OGC::Service::XMLWriter::Streaming($responder, $headers, $declaration);
 
 Using $w as XMLWriter sets writer, which is obtained from $responder,
 to write XML. The writer is closed when $w is destroyed.
 
-$content_type and $declaration are optional. The defaults are
-'text/xml; charset=utf-8' and '<?xml version="1.0"
+$headers and $declaration are optional. The defaults are
+'Content-Type' => 'text/xml; charset=utf-8' and '<?xml version="1.0"
 encoding="UTF-8"?>'.
 
 =cut
@@ -612,9 +700,15 @@ use Modern::Perl;
 our @ISA = qw(Geo::OGC::Service::XMLWriter Plack::Util::Prototype); # can't use parent since Plack is not yet
 
 sub new {
-    my ($class, $responder, $content_type, $declaration) = @_;
-    $content_type //= 'text/xml; charset=utf-8';
-    my $self = $responder->([200, [ 'Content-Type' => $content_type ]]);
+    my ($class, $responder, $headers, $declaration) = @_;
+    my %headers;
+    if (ref $headers) {
+        %headers = @$headers;
+    } else {
+        $headers{'Content-Type'} = $headers;
+    }
+    $headers{'Content-Type'} //= 'text/xml; charset=utf-8';
+    my $self = $responder->([200, [%headers]]);
     $self->{declaration} = $declaration //= '<?xml version="1.0" encoding="UTF-8"?>';
     return bless $self, $class;
 }
@@ -637,13 +731,13 @@ A helper class for writing XML into a cache.
 
 =head2 SYNOPSIS
 
- my $w = Geo::OGC::Service::XMLWriter::Caching($content_type, $declaration);
+ my $w = Geo::OGC::Service::XMLWriter::Caching($headers, $declaration);
  $w->stream($responder);
 
 Using $w to produce XML caches the XML. The cached XML can be
 written by a writer obtained from a $responder.
 
-$content_type and $declaration are optional. The defaults are as in
+$headers and $declaration are optional. The defaults are as in
 Geo::OGC::Service::XMLWriter::Streaming.
 
 =cut
@@ -654,10 +748,17 @@ use Modern::Perl;
 our @ISA = qw(Geo::OGC::Service::XMLWriter);
 
 sub new {
-    my ($class, $content_type, $declaration) = @_;
+    my ($class, $headers, $declaration) = @_;
+    my %headers;
+    if (ref $headers) {
+        %headers = @$headers;
+    } else {
+        $headers{'Content-Type'} = $headers;
+    }
+    $headers{'Content-Type'} //= 'text/xml; charset=utf-8';
     my $self = {
         cache => [],
-        content_type => $content_type //= 'text/xml; charset=utf-8',
+        headers => \%headers,
         declaration => $declaration //= '<?xml version="1.0" encoding="UTF-8"?>'
     };
     $self->{cache} = [];
@@ -683,7 +784,7 @@ sub stream {
     my $self = shift;
     my $responder = shift;
     my $debug = shift;
-    my $writer = $responder->([200, [ 'Content-Type' => $self->{content_type} ]]);
+    my $writer = $responder->([200, [ %{$self->{headers}} ]]);
     $writer->write($self->{declaration});
     my $xml = '';
     for my $line (@{$self->{cache}}) {
